unit uModeloGenERNC;

{$mode delphi}

interface

uses
  Classes, SysUtils, Math, xmatdefs, matreal, matent, uauxiliares,
  umodelosintcegh, uro_aire;

type

  // El modelo de Parque Eólico se obtiene a partir de muestras diezminutales de una serie de viento
  // diezminutal y de las series de potencia generada diezminutal.
  // (NPuntosD): Cantidad de sectores para el modelo.Se calcula
  // un modelo para cada sector con centro en al dirección 360/NPuntosD * (k-1) con k= 1 .. NPuntosD
  // (SolapamientoD): Las muestras que se consideran en el cálculo del modelo de sector "k" serán
  // aquellas cuyo ánguo de dirección está comprendido entre en centro del sector más menos
  // el valor (360/NPuntosD)/2 * (1+SolapamientoD).
  // Ejemplo de uso: NPUntosD = 24 , Solapamiento=1 implica que se crearán 24 modelos para direcciones
  // cada 15° y que para el modelo "k" se considerarán las muestras con dirección entre (k-1)*24 - 15
  // y (k-1)*24 + 15.


  TModeloParqueEolico = class
    nombre: string; // nombre del generador
    base_PV: NReal; // ro_ v^3 tomada com base
    base_PE: NReal; // MW tomados como base.

    NPuntosD: integer; // Cantidad de Puntos de Dirección.
    OverlappingD: NReal; // solapamiento

    NPuntosP: integer; // Cantidad de Puntos en las curvas de potencias.
    PV: TDAOfVectR; // Potencias del Viento en p.u.
    PE: TDAOfVectR; // Potencias Eleéctricas en p.u.
    cntMuestras: TDAOfVectE;


    // Crea una estructura vacía.
    constructor Create(nombre: string; NPuntosD: integer;
      OverlappingD: NReal; NPuntosP: integer; base_PE: NReal);

    // Destruye la instancia.
    procedure Free;


    // Cargar series de medidas para calcular el modelo del parque.
    procedure LoadSeries(archi: string);

    // Cargar series de vientos para estimar potencia generada.
    procedure LoadVientos(archi: string);

    // Cargar series de potencia estimada.
    procedure LoadEstimacion(archi: string);

    // Calcula los parámetros del modelo a partir de un archivo de muestras.
    procedure CalcularFromMuestras;

    // Guarda el modelo en disco
    procedure WriteToArchi(archi: string);

    // Carga el modelo desde disco.
    constructor Create_ReadFromArchi(archi: string);

    // estima la potencia en función de los parámetros y el modelo cargado
    function CalculaPotencia(Vel_mps, Dir_gr, Temp_grC, Pre_hPa,
      HumRel_p100: NReal): TVectR;


    //estima la potencia eolica generada a partir de las medidas de viento de la Estación Meteo. del parque.
    procedure EstimarPotenciaEolica;

    //estima la disponibilidad para una determinada ventana móvil de horas.
    procedure EstimarDisponibilidad(Ven1, Ven2, Ven3, Ven4: integer);


  private
    series, series_a, series_b, series3, series4, series5: TSeriesDeDatos;
    dang: NReal;

  end;

implementation



// Crea una estructura vacía.
constructor TModeloParqueEolico.Create(nombre: string; NPuntosD: integer;
  OverlappingD: NReal; NPuntosP: integer; base_PE: NReal);
var
  k: integer;
begin
  inherited Create;
  dang := 360.0 / NPuntosD;
  self.Nombre := nombre;
  // base_PV:= ro_aire( Base_Temperatura_grC, Base_Presion_hPA, Base_HumedadRelativa_p100 ) * power( Base_Velocidad_mps, 3 );
  base_PV := 1;
  self.base_PE := base_PE;
  self.NPuntosD := NPuntosD;
  self.OverlappingD := OverlappingD;
  self.NPuntosP := NPuntosP; // Cantidad de Puntos en las curvas de potencias.
  setlength(PV, NPuntosD);
  setlength(PE, NPuntosD);
  setlength(cntMuestras, NPuntosD);
  for k := 0 to NPuntosD - 1 do
  begin
    PV[k] := TVectR.Create_Init(NPUntosP);
    PE[k] := TVectR.Create_Init(NPUntosP);
    cntMuestras[k] := TVectE.Create_Init(nPuntosP);
  end;
  series := nil;
end;


procedure TModeloParqueEolico.Free;
var
  k: integer;
begin
  for k := 0 to NPuntosD - 1 do
  begin
    PE[k].Free;
    PV[k].Free;
    cntMuestras[k].Free;
  end;
  setlength(PE, 0);
  setlength(PV, 0);
  setlength(cntMuestras, 0);
  if series <> nil then
  begin
    series.Free;
    series := nil;
  end;
  inherited Free;
end;

procedure TModeloParqueEolico.WriteToArchi(archi: string);
var
  f: textfile;
  k, j: integer;

begin
  assignfile(f, archi);
  rewrite(f);
  writeln(f, nombre);
  writeln(f, base_PE);
  writeln(f, base_PV);
  writeln(f, NPuntosD);
  writeln(f, OverlappingD);
  writeln(f, NPUntosP);
  writeln(f, ' --- potencias de viento --- ');
  for k := 0 to NPuntosD - 1 do
  begin
    Write(f, k);
    for j := 0 to NPuntosP do
      Write(f, #9, PV[k].e(j));
    writeln(f);
  end;
  writeln(f, ' --- potencias de parque --- ');
  for k := 0 to NPuntosD - 1 do
  begin
    Write(f, k);
    for j := 0 to NPuntosP do
      Write(f, #9, PE[k].e(j));
    writeln(f);
  end;
  writeln(f, ' --- cantidad de muestras --- ');
  for k := 0 to NPuntosD - 1 do
  begin
    Write(f, k);
    for j := 0 to NPuntosP do
      Write(f, #9, cntMuestras[k].e(j));
    writeln(f);
  end;
  closefile(f);
end;

// Carga el modelo desde disco.
constructor TModeloParqueEolico.Create_ReadFromArchi(archi: string);
var
  f: textfile;
  k, j: integer;
  base_PV_: NReal;
  k_: integer;
  s: string;
begin
  assignfile(f, archi);
  reset(f);
  readln(f, nombre);
  readln(f, base_PE );
  readln(f, base_PV_);
  readln(f, NPuntosD);
  readln(f, OverlappingD);
  readln(f, NPUntosP);
  self.Create(Nombre, NpuntosD, OverlappingD, NPUntosP, base_PE);
  base_PV := base_PV_;
  readln(f, s);
  for k := 0 to NPuntosD - 1 do
  begin
    Read(f, k_);
    for j := 0 to NPuntosP - 1 do
      Read(f, PV[k].pv[j]);
    readln(f);
  end;
  readln(f, s);
  for k := 0 to NPuntosD - 1 do
  begin
    Read(f, k_);
    for j := 0 to NPuntosP - 1 do
      Read(f, PE[k].pv[j]);
    readln(f);
  end;
  readln(f, s);
  for k := 0 to NPuntosD - 1 do
  begin
    Read(f, k_);
    for j := 0 to NPuntosP - 1 do
      Read(f, cntMuestras[k].pv[j]);
    readln(f);
  end;
  closefile(f);
end;


// estima la potencia en función de los parámetros y el modelo cargado. Como salida publica la
//potencia del viento de entrada y la estimación de la potencia generada por el parque.
function TModeloParqueEolico.CalculaPotencia(
  Vel_mps, Dir_gr, Temp_grC, Pre_hPa, HumRel_p100: NReal): TVectR;
var
  rang_a, rang_b, indice_a, indice_b, fro, pot_viento_muestra,
  PEinterpol: NReal;
  angulo_a, angulo_b: string;
  pot_viento_modelo_a, pot_viento_modelo_b, pot_electrica_modelo_a,
  pot_electrica_modelo_b, salida: TVectR;

begin

  rang_a := (Dir_gr / dang);
  rang_b := ((Dir_gr + dang) / dang);
  salida := TVectR.Create_Init(2);

  //leo el archivo del modelo correspondiente al angulo rang//

  angulo_a := IntToStr(trunc(rang_a) * trunc(dang * 100));
  angulo_b := IntToStr(trunc(rang_b) * trunc(dang * 100));

  if angulo_a = '36000' then
    angulo_a := '0';
  if angulo_b = '36000' then
    angulo_b := '0';

  series_a := TSeriesDeDatos.CreateFromArchi(
    'c:\basura\Modelo_Parque\resdir_' + Nombre + '_' + angulo_a + '.xlt');
  series_b := TSeriesDeDatos.CreateFromArchi(
    'c:\basura\Modelo_Parque\resdir_' + Nombre + '_' + angulo_b + '.xlt');

  pot_viento_modelo_a := series_a.series[0];
  pot_electrica_modelo_a := series_a.series[2];

  pot_viento_modelo_b := series_b.series[0];
  pot_electrica_modelo_b := series_b.series[2];

  //calculo la potencia de la muestra de viento //
  fro := ro_aire(Temp_grC, Pre_hPa, HumRel_p100) / base_PV;
  pot_viento_muestra := fro * power(Vel_mps, 3.0);
  salida.pon_e(1, pot_viento_muestra);

  // estimo la potencia del parque en función de la potencia de la muestra de viento


  indice_a := pot_viento_modelo_a.inv_interpol(pot_viento_muestra);
  indice_b := pot_viento_modelo_b.inv_interpol(pot_viento_muestra);

  //aux1:=pot_electrica_modelo_a.interpol(indice_a);
  //aux2:=pot_electrica_modelo_b.interpol(indice_b);

  PEinterpol := (1 - (rang_a - trunc(rang_a))) * pot_electrica_modelo_a.interpol(indice_a) +
    (rang_a - trunc(rang_a)) * pot_electrica_modelo_b.interpol(indice_b);

  salida.pon_e(2, PEinterpol);

  Result := salida;

end;

procedure TModeloParqueEolico.LoadVientos(archi: string);
begin
  series3 := TSeriesDeDatos.CreateFromArchi(archi);
end;


//a partir de un archivo con medidas de viento de la estación met estima la potencia generada
//por el parque cargado en el modelo.
procedure TModeloParqueEolico.EstimarPotenciaEolica;
var
  Pot_estimada: TVectR;
  k: integer;
  cont: integer;
  f: textfile;
  sPot_MW, sDir_gr, sVel_mps, sTemp_grC, sPre_hPa, sHumRel_p100: TVectR;

begin

  assignfile(f, 'c:\basura\Modelo_Parque\Potencia_Estimada_' + Nombre + '_' + '.xlt');
  rewrite(f);
  writeln(f, IntToStr(3), #9, 'NSeries');
  writeln(f, IntToStr(2015), #9, IntToStr(01), #9, IntToStr(01), #9,
    IntToStr(0), #9, IntToStr(0), #9, IntToStr(0),
    #9'//anio mes dia hora minuto segundo fecha de la primera muestra');
  writeln(f, IntToStr(0), #9, '//Periodo de muestreo en horas');
  writeln(f, IntToStr(series3.NPuntos), #9, 'NPuntos');
  writeln(f, IntToStr(0), #9, 'NPuntos por ciclo');

  writeln(f, ''#9'Velocidad del viento'#9'Potencia del viento EM [MW]'#9'Potencia estimada del Parque [MW]');

  sPot_MW := series3.series[0];
  sDir_gr := series3.series[1];
  sVel_mps := series3.series[3];

  sTemp_grC := series3.series[4];
  sPre_hPa := series3.series[5];
  sHumRel_p100 := series3.series[6];

  cont := 1;
  Pot_estimada := TVectR.Create_init(2);

  for k := 1 to series3.NPuntos do

  begin

    Pot_estimada := CalculaPotencia(sVel_mps.e(k), sDir_gr.e(k),
      sTemp_grC.e(k), sPre_hPa.e(k), sHumRel_p100.e(k));

    writeln(f, k, #9, sVel_mps.e(k), #9, Pot_estimada.e(1), #9, Pot_estimada.e(2));
    if k = cont * trunc(series3.NPuntos / 10) then
    begin
      writeln('va ' + IntToStr(cont * 10) + '%');
      cont := cont + 1;
    end;

  end;
  closefile(f);
end;


procedure TModeloParqueEolico.LoadSeries(archi: string);
begin
  series := TSeriesDeDatos.CreateFromArchi(archi);
end;

procedure TModeloParqueEolico.CalcularFromMuestras;
var
  Pot_MW, froV3: TVectR;
  k: integer;
  cnt: integer;
  sPot_MW, sDir_gr, sVel_mps, sTemp_grC, sPre_hPa, sHumRel_p100: TVectR;
  fro: NReal;
  k_dir: integer;
  dir_sel, dir: NReal;

  f, fapro: textfile;

  x, y: NReal;

  EAcumCluster: TVectR;
  EnergiaTotalDelCluster: NReal;
  PotVal: NReal;

  idx: TDAOfNInt;
  frov3_x, Pot_MW_y: TVectR;
  cnts: TVectE;
  j_ant, j_j, j_sig: integer;
  dE: NReal;
  x_ant, x_j, x_sig: NReal;
  E_ant, E_j, E_sig: NReal;

  kr: NReal;

  dang_overlapping: NReal;
begin

  sPot_MW := series.series[0];
  sDir_gr := series.series[1];
  sVel_mps := series.series[3];

  sTemp_grC := series.series[4];
  sPre_hPa := series.series[5];
  sHumRel_p100 := series.series[6];


  assignfile(fapro, 'c:\basura\Modelo_Parque\apros_' + Nombre + '.xlt');
  rewrite(fapro);
  //  writeln(fapro, 'dir', #9, 'ra1', #9, 'rb1', #9, 'rxc', #9, 'ra2', #9, 'rb2');

  Write(fapro, 'dir_sel');
  for k := 1 to NPuntosP do
    Write(fapro, #9, 'frov3_' + IntToStr(k));
  for k := 1 to NPuntosP do
    Write(fapro, #9, 'Pot_MW_' + IntToStr(k));
  for k := 1 to NPuntosP do
    Write(fapro, #9, 'cnts_' + IntToStr(k));
  writeln(fapro);


  dang := 360.0 / NPuntosD;
  dang_overlapping := dang / 2.0 * (1 + OverlappingD);

  for k_dir := 0 to NPuntosD - 1 do
  begin
    dir_sel := k_dir * dang;

    Pot_MW := TVectR.Create_Init(series.NPuntos);
    froV3 := TVectR.Create_Init(series.NPuntos);

    // filtramos por dirección
    cnt := 0;
    for k := 1 to series.NPuntos do
    begin
      dir := sDir_gr.e(k);
      if ((abs(dir - dir_sel) < dang_overlapping) or
        (abs(dir - 360 - dir_sel) < dang_overlapping) or
        (abs(dir + 360 - dir_sel) < dang_overlapping)) then
      begin
        Inc(cnt);
        fro := ro_aire(sTemp_grC.e(k), sPre_hPa.e(k), sHumRel_p100.e(k)) / base_PV;
        froV3.pon_e(cnt, fro * power(sVel_mps.e(k), 3.0));

        potVal := sPot_MW.e(k) / base_PE;
        Pot_MW.pon_e(cnt, potVal);
      end;
    end;

    frov3.resize(cnt);
    Pot_MW.resize(cnt);

    setlength(idx, cnt + 1);
    for k := 0 to high(idx) do
      idx[k] := k;

    frov3.Sort_idx(True, idx);

    EacumCluster := TVectR.Create_init(pot_MW.n + 1);

    // inpongo el punto v= 0 -> Eacum = 0
    EnergiaTotalDelCluster := 0.0;
    EAcumCluster.pon_e(1, 0.0);

    for k := 1 to pot_MW.n do
    begin
      potval := pot_MW.e(idx[k]);
      EAcumCluster.pon_e(k + 1, EAcumCluster.e(k) + potval);
      EnergiaTotalDelCluster := EnergiaTotalDelCluster + potval;
    end;

    // Apuntamos el modelo de la dirección k_Dir
    frov3_x := PV[k_Dir];
    pot_MW_y := PE[k_Dir];
    cnts := cntMuestras[k_Dir];

    dE := (EnergiaTotalDelCluster - 3.0) / (NPuntosP);

    // Primer punto
    j_ant := 1;
    x_ant := 0;
    E_ant := 0;

    j_j := trunc((EAcumCluster.inv_interpol(
      3.0 * 0.9) + EAcumCluster.inv_interpol(3.0 * 1.1)) / 2.0 + 0.5);
    j_j := max(1, min(j_j, EAcumCluster.n));
    x_j := froV3.e(j_j - 1);
    E_j := EAcumCluster.e(j_j);

    k := 1;
    j_sig := trunc((EAcumCluster.inv_interpol(3.0 + k * dE * 0.9) +
      EAcumCluster.inv_interpol(3.0 + k * dE * 1.1)) / 2.0 + 0.5);
    j_sig := max(1, min(j_sig, EAcumCluster.n - 1));
    x_sig := froV3.e(j_sig - 1);
    E_sig := EAcumCluster.e(j_sig);

    froV3_x.pon_e(1, x_j);
    pot_MW_y.pon_e(1, 0); //(E_sig - 0)/( j_sig- 1 ));
    cnts.pon_e(1, j_j - 1);   //  /?????

    for k := 2 to NPuntosP - 1 do
    begin
      x_ant := x_j;
      x_j := x_sig;
      j_ant := j_j;
      j_j := j_sig;
      E_ant := E_j;
      E_j := E_sig;
      j_sig := trunc((EAcumCluster.inv_interpol(3 + k * dE * 0.9) + EAcumCluster.inv_interpol(
        3 + k * dE * 1.1)) / 2.0 + 0.5);
      j_sig := max(1, min(j_sig, EAcumCluster.n - 1));
      E_sig := EAcumCluster.e(j_sig);
      x_sig := froV3.e(j_sig - 1);
      frov3_x.pon_e(k, x_j);
      if (E_sig - E_ant) = 0 then
        pot_MW_y.pon_e(k, pot_MW_y.e(k - 1))
      else
        pot_MW_y.pon_e(k, (E_sig - E_ant) / (j_sig - j_ant));
      cnts.pon_e(k, j_sig - j_ant);
    end;

    // Ultimo punto
    {x_ant:= x_j; x_j:= x_sig;
    j_ant:= j_j;  j_j:= j_sig;
    E_ant:= E_j; E_j:= E_sig;

    j_sig:= EAcumCluster.n;
    x_sig:=frov3.e( j_sig-1 );
    E_sig:= EAcumCluster.e( j_sig );

    frov3_x.pon_e(NPuntosP, (x_ant+x_sig)/2.0 );
    pot_MW_y.pon_e( NPuntosP, ( E_sig - E_ant )/( j_sig- j_ant ) );
    cnts.pon_e( NPuntosP, j_sig- j_ant );
     }

    Write(fapro, dir_sel);
    for k := 1 to NPuntosP do
      Write(fapro, #9, frov3_x.e(k));
    for k := 1 to NPuntosP do
      Write(fapro, #9, Pot_MW_y.e(k));
    for k := 1 to NPuntosP do
      Write(fapro, #9, cnts.e(k));
    writeln(fapro);



    assignfile(f, 'c:\basura\Modelo_Parque\resdir_' + Nombre + '_' +
      IntToStr(trunc(k_dir * dang * 100)) + '.xlt');
    rewrite(f);
    writeln(f, IntToStr(3), #9, 'NSeries');
    writeln(f, IntToStr(2015), #9, IntToStr(01), #9, IntToStr(01), #9,
      IntToStr(0), #9, IntToStr(0), #9, IntToStr(0),
      #9'//anio mes dia hora minuto segundo fecha de la primera muestra');
    writeln(f, IntToStr(0), #9, '//Periodo de muestreo en horas');
    writeln(f, cnt, #9, 'NPuntos');
    writeln(f, IntToStr(0), #9, 'NPuntos por ciclo');

    writeln(f, ''#9'frov3'#9'P[MW]'#9'Px[MW]');

    for k := 1 to cnt do
    begin
      x := frov3.e(k);
      kr := frov3_x.inv_interpol(x);
      if kr < 1 then
        y := 0
      else
        y := Pot_MW_y.interpol(kr);
      writeln(f, k, #9, x, #9, Pot_MW.e(idx[k]), #9, y);
    end;
    closefile(f);

    EAcumCluster.Free;

    frov3.Free;
    Pot_MW.Free;
    setlength(idx, 0);

  end;
  closefile(fapro);
end;

procedure TModeloParqueEolico.LoadEstimacion(archi: string);
begin
  series4 := TSeriesDeDatos.CreateFromArchi(archi);
end;


procedure TModeloParqueEolico.EstimarDisponibilidad(Ven1, Ven2, Ven3, Ven4: integer);

var
  Pot_Real, Pot_Estimada, Vel_Viento, Factor, Fnorm, PotPM1, PotPM2, PotPM3, PotPM4: TVectR;
  fle: textfile;
  largo, k, m, j, z, Venmax: integer;
  aux1, aux2, aux3, aux4: NReal;

begin

  //Ventana:=10;
  Vel_Viento := series3.series[3];
  Pot_Real := series3.series[0];
  //series4 := TSeriesDeDatos.CreateFromArchi( 'c:\basura\Modelo_Parque\Potencia_Estimada_' + Nombre + '_' + '.xlt' );
  Pot_Estimada := series4.series[2];
  largo := series4.NPuntos;

  Factor := TVectR.Create_Init(largo);
  Fnorm := TVectR.Create_Init(largo);
  //aux1:=TVectR.Create_Init(largo);

  PotPM1 := TVectR.Create_Init(largo);
  PotPM2 := TVectR.Create_Init(largo);
  PotPM3 := TVectR.Create_Init(largo);
  PotPM4 := TVectR.Create_Init(largo);

  Venmax := max(Ven1, Ven2);
  Venmax := max(Venmax, Ven3);
  Venmax := max(Venmax, Ven4);

  assignfile(fle, 'c:\basura\Modelo_Parque\Disponibilidad_' + Nombre + '_' + '.xlt');
  rewrite(fle);
  writeln(fle, IntToStr(3), #9, 'NSeries');
  writeln(fle, IntToStr(2015), #9, IntToStr(01), #9, IntToStr(01),
    #9, IntToStr(0), #9, IntToStr(0), #9, IntToStr(0),
    #9'//anio mes dia hora minuto segundo fecha de la primera muestra');
  writeln(fle, IntToStr(0), #9, '//Periodo de muestreo en horas');
  writeln(fle, IntToStr(largo - Venmax - 1), #9, 'NPuntos');
  writeln(fle, IntToStr(0), #9, 'NPuntos por ciclo');

  writeln(fle, ''#9'Velocidad del viento'#9'Potencia real del Parque [MW]'#9'Potencia estimada del Parque [MW]'#9'Pot real / estimada'#9'Promedio Movil ' +
    IntToStr(Ven1), #9'Promedio Movil ' + IntToStr(Ven2), #9'Promedio Movil ' +
    IntToStr(Ven3), #9'Promedio Movil ' + IntToStr(Ven4), #9'Disp');


  for k := 1 to largo do
  begin

    Factor.pon_e(k, Pot_Real.e(k) / (Pot_Estimada.e(k) + 0.1));

  end;

  for z := 1 to Ven1 do
    PotPM1.pon_e(z, 0);

  for z := 1 to Ven2 do
    PotPM2.pon_e(z, 0);

  for z := 1 to Ven3 do
    PotPM3.pon_e(z, 0);

  for z := 1 to Ven4 do
  begin
    PotPM4.pon_e(z, 0);
    Fnorm.pon_e(z, 0);
  end;

  m := 0;
  while m <= (largo - Venmax - 1) do
  begin

    aux1 := 0;
    aux2 := 0;
    aux3 := 0;
    aux4 := 0;

    for j := 1 to Ven1 do

      aux1 := aux1 + Factor.e(m + j);
    aux1 := aux1 / Ven1;

    for j := 1 to Ven2 do

      aux2 := aux2 + Factor.e(m + j);
    aux2 := aux2 / Ven2;

    for j := 1 to Ven3 do

      aux3 := aux3 + Factor.e(m + j);
    aux3 := aux3 / Ven3;

    for j := 1 to Ven4 do

      aux4 := aux4 + Factor.e(m + j);
    aux4 := aux4 / Ven4;


    PotPM1.pon_e(m + Ven1 + 1, aux1);
    PotPM2.pon_e(m + Ven2 + 1, aux2);
    PotPM3.pon_e(m + Ven3 + 1, aux3);
    PotPM4.pon_e(m + Ven4 + 1, aux4);

    if PotPM4.e(m + Ven4 + 1) < 1 then
      Fnorm.pon_e(m + Ven4 + 1, 1)
    else
      Fnorm.pon_e(m + Ven4 + 1, 0);

    m := m + 1;

    writeln(fle, m, #9, Vel_viento.e(m), #9, Pot_Real.e(m), #9, Pot_estimada.e(m),
      #9, Factor.e(m), #9, PotPM1.e(m), #9, PotPM2.e(m), #9, PotPM3.e(m), #9, PotPM4.e(m),
      #9, Fnorm.e(m));

  end;
  closefile(fle);

end;

end.
