unit uHidroDePasada;

interface

uses
  Classes,
  Math, xMatDefs, uSimplex,
  ucosa, uGlobs, uFechas, uNodos,
  SysUtils,
  uFichasLPD,
  uGeneradores,
  uconstantesSimSEE,
  uFuentesAleatorias;

resourcestring
  rsGHidraulicoPasada = 'Generador hidráulico de pasada';

const
  ro_Agua = 1000; // kg/m3 densidad del agua
  g_Gravitacion = 9.8; // N/kg constante gravitatoria

type
  THidroDePasada = class; // forward declaration

{Las cotas están medidas desde algún nivel (por ejemplo el del mar).
Para poder calcular el salto se suministra también la cota de aguas abajo
o de descarga de la turbina referida al mismo nivel. (hDescarga [m])
El nivel de descarga puede ser especificado dando su valor con la variable
hDescarga o mediante la referencia a otra central en cuyo caso se considera
que el nivel de descarga es el del lago de la otra central más el valor
hDescarga. En este caso la variable hDescarga se utiliza para tener en
cuenta cambios en la referencia de niven entre una central y otra.
Si todas las cotas están referidas al nivel del mar no hay cambios de
referencias y hDescarga se pondrá en cero.


Pérdida de Salto por Caudal Erogado.
====================================
El caudal erogado QErogado [m3/s] es calculado al final de cada paso
de simulación como la suma del caudal vertido más el caudal turbinado.
Dicho caudal puede afectar el nivel de aguas abajo reduciendo el salto
efectivo en la central. Para tener en cuenta este efecto se usan los
parámetros: caQE y cbQE calculados para modelar la pérdida de salto
de acuerdo a la siguiente ecuación:

dh(QE) = caQE* QE + cbQE* QE^2
*)}

  { TFichaHidroDePasada }

  TFichaHidroDePasada = class(TFichaLPD)
  public

    (**************************************************************************)
    (*              A T R I B U T O S   P E R S I S T E N T E S               *)
    (**************************************************************************)

    // Si la fuente en lugar de caudal en m3/s es de escurrimiento
    // en mm/mes hay que especificar el área de la cuenca a considerar
    AreaCuenca_ha: NReal;
    cv_agua_USD_Hm3: NReal; // valor del agua en USD/Hm3
    hDescarga: NReal; //[m] cota de la descarga para cálculo del salto
    hToma: NReal; //[m] cota de la toma, la altura efectiva sera hToma - hDescarga
    // Coeficientes de afectación del salto por caudal erogado
    // dh(QE) = caQE* QE + cbQE* QE^2
    caQE: NReal;
    cbQE: NReal;
    ren: NReal; //[pu]= 0.95; // rendimiento complexivo de turbina y generador
    Pmax_Gen: NReal; //[MW]= 240; // Potencia maxima hidraulica
    Qmax_Turb: NReal; //[m3/s]
    SaltoMinimoOperativo: NReal; // salto mínimo para funcionamiento de las turbinas
    fDispo: NReal; //[pu]  factor de disponibilidad fortuito
    tRepHoras: NReal;
    central_lagoDescarga: TGeneradorHidraulico;

         (** DATOS CENTRALES URUGUAY
         BONETE= Dr. Gabriel Terra
         =========================
         hmax= 86.5
         hmin= 70
         Vmax= 1.667E+16
         Vhmed= 8.335E+15
         hDescarga= 54.5394
         caQE= 4.689E-04
         cbQE= 5.077E-07

         ren= 0.85 ??? ojo ver mejor este número
         Pmax_Gen= 38.8
         Qmax_Turb= 170.0
         fDispo= 0.995

         *)
    centralesAguasArriba: TListaCentralesAguasArriba{of TGeneradorHidraulico};

    HayRestriccionEmaxPasoDeTiempo: boolean; // indica si se aplica la restricción
    EmaxPasoDeTiempo: NReal; // Energía maxima generable en un paso de tiempo
    // Pagos a los generadores
    PagoPorDisponibilidad_USD_MWh: NReal; // [USD/MWh] Pago por Potencia
    PagoPorEnergia_USD_MWh: NReal; // [USD/MWh] Pago por Potencia

    (**************************************************************************)



    constructor Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      AreaCuenca_ha, hDescarga, hToma: NReal; cv_agua_USD_Hm3: NReal;
      central_lagoDescarga: TGeneradorHidraulico;
      centralesAguasArriba: TListaCentralesAguasArriba;
      caQE, cbQE, ren_, Pmax_Gen_, Qmax_Turb_: NReal; SaltoMinimoOperativo: NReal;
      fDispo_: NReal; tRepHoras: NReal; HayRestriccionEmaxPasoDeTiempo: boolean;
      EmaxPasoDeTiempo: NReal; PagoPorDisponibilidad_USD_MWh: NReal;
      PagoPorEnergia_USD_MWh: NReal);
    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    {$ENDIF}
    function infoAd_: string; override;
    procedure Free; override;

    {$IFDEF MIGRACION_PERSISTENCIA}
    class function CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
      xVersion: integer = -2): TDataColumnListOfCosa; override;
    {$ENDIF}
  published
  {$IFDEF MIGRACION_PERSISTENCIA}
    property _AreaCuenca_ha: NReal
      read AreaCuenca_ha write AreaCuenca_ha;
    property _cv_agua_USD_Hm3: NReal
      read cv_agua_USD_Hm3 write cv_agua_USD_Hm3;
    property _hDescarga: NReal
      read hDescarga write hDescarga;
    property _hToma: NReal
      read hToma write hToma;
    property _caQE: NReal
      read caQE write caQE;
    property _cbQE: NReal
      read cbQE write cbQE;
    property _ren: NReal
      read ren write ren;
    property _Pmax_Gen: NReal
      read Pmax_Gen write Pmax_Gen;
    property _Qmax_Turb: NReal
      read Qmax_Turb write Qmax_Turb;
    property _SaltoMinimoOperativo: NReal
      read SaltoMinimoOperativo write SaltoMinimoOperativo;
    property _fDispo: NReal
      read fDispo write fDispo;
    property _tRepHoras: NReal
      read tRepHoras write tRepHoras;
    property _central_lagoDescarga: TGeneradorHidraulico
      read central_lagoDescarga write central_lagoDescarga;
    property _centralesAguasArriba: TListaCentralesAguasArriba
      read centralesAguasArriba write centralesAguasArriba;
    property _HayRestriccionEmaxPasoDeTiempo: boolean
      read HayRestriccionEmaxPasoDeTiempo write HayRestriccionEmaxPasoDeTiempo;
    property _EmaxPasoDeTiempo: NReal
      read EmaxPasoDeTiempo write EmaxPasoDeTiempo;
    property _PagoPorDisponibilidad_USD_MWh: NReal
      read PagoPorDisponibilidad_USD_MWh write PagoPorDisponibilidad_USD_MWh;
    property _PagoPorEnergia_USD_MWh: NReal
      read PagoPorEnergia_USD_MWh write PagoPorEnergia_USD_MWh;
  {$ENDIF}

  end;


  { THidroDePasada }

  THidroDePasada = class(TGeneradorHidraulico)
  public

    (**************************************************************************)
    (*              A T R I B U T O S   P E R S I S T E N T E S               *)
    (**************************************************************************)

    fuenteDeAportes: TFuenteAleatoria;
    nombreBorne: string;
    // Si es TRUE, la fuente suministra el escurrimiento en mm/mes
    // y para calcular el caudal hay que multiplicar por el área
    // y dividir por los segundos del mes.  (mm/mes) /1000 * Area[km2]*1000*1000 /(730*3600)
    // si es false, la fuente
    // suministra el caudal medio en m3/s.
    flg_FuenteDeEscurrimientos: boolean;
    flg_IgualPotenciaEnTodosLosPostes: boolean;

    (**************************************************************************)

    pa: TFichaHidroDePasada;

    numeroBorne: integer;

    //----------- variables auxiliares
    ro_g_ren: NReal; // ro_Agua * g_gravitacion * ren /(1E6*3600)
    VTmin: NReal; // volumen turbinado mínimo en el paso para cumplir con QTmin

    // variables calculadas por paso de tiempo durante la simulación
    h_actual: NReal;
    Salto: NReal; // salto actual efectivo

    QErogado_IterAnterior: NReal;
    // Usamos esta variable para almacenar el QErogado en la
    // iteración anterior. Esto se hace en leerResultadosDelPaso
    // antes de cargar el nuevo valor de QErogado.


    PmaxParaQmax: NReal; // Pontecia máxima por máquina considerando Qmax y la cota
    NMaquinasDisponibles: integer;
    PDisponible_Central: NReal; // Disponibilidad MW para remuneración.
    Pmax_Central: NReal; // [MW] Potencia Máxima disponible en la central
    Pmax_Gen: NReal; // [MW] Potencia máxima generable por turbina-generador
    Qmax: NReal; // [m3/s] Caudal máximo turbinable por turbina-generador
    PuedoGenerar: boolean;
    // indica si la Central está en condiciones de ofrecer potencia o no
    cv_USD_MWh: NReal; // valor del Agua expresado en USD/MWh
    cv_agua_USD_Hm3_MLRB: NReal;
    // multiplicador de Lagrange de la restricción de balance.
    dh_RedQE: NReal; // [m] pérdida de salto por caulda erogado.

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
      flg_CalcularGradienteDeInversion: boolean; fuenteDeAportes: TFuenteAleatoria;
      nombreBorne: string; flg_FuenteDeEscurrimientos,
      flg_IgualPotenciaEnTodosLosPostes: boolean; TonCO2xMWh: NReal;
      LowCostMustRun, CleanDevelopmentMechanism: boolean);

    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    {$ENDIF}
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure RegistrarParametrosDinamicos; override;

    function PotenciaFirme: NReal; override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    //      procedure Sim_Inicio; override;
    procedure Sim_Cronica_Inicio; override;
    procedure SorteosDelPaso(sortear: boolean); override;
    procedure PrepararPaso_ps; override;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;

    procedure opt_cargue(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;
{$IFDEF SPXCONLOG}
    procedure spx_NombrarVariables(s: TSimplex); override;
{$ENDIF}
    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure optx_nvxs(var ixr, ixd, iauxr, iauxd: integer); override;
    procedure PosicionarseEnEstrellita; override;
    procedure AcumAux1(peso: NReal); override;
    procedure SetAux1; override;

    function opt_necesitoIterar(kIteracion: integer; var errRelativo: NReal): boolean;
      override;

    procedure Free; override;
    function get_pa_FD(kTipoUnidad: integer): NReal; override;
    function get_pa_TMR(kTipoUnidad: integer): NReal; override;

    procedure CambioFichaPD; override;

    procedure PubliVars; override;

    function CotaAguasArriba: NReal; override;
    // Calcula la cota de la descarga para poder calcular el Salto
    function CotaDescarga: NReal; override;
    // Caclula la reducción del salto causada por el caudal Erogado
    function ReduccionDeCotaPorCaudalErogado(QErogado: NReal): NReal; override;
    function centralLagoDescarga: TGeneradorHidraulico; override;

    procedure ResolverEncadenamientos(s: TSimplex); override;

    procedure dump_Variables(var f: TextFile; charIndentacion: char); override;
    function vertimientoPorPoste: boolean; override;
    function NVariablesPotenica: integer; override;


    {$IFDEF MIGRACION_PERSISTENCIA}
    class function CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
      xVersion: integer = -2): TDataColumnListOfCosa; override;
    {$ENDIF}
    procedure AfterInstantiation(); override;

  published
  {$IFDEF MIGRACION_PERSISTENCIA}
    property _fuenteDeAportes: TFuenteAleatoria
      read fuenteDeAportes write fuenteDeAportes;
    property _nombreBorne: string
      read nombreBorne write nombreBorne;
    property _flg_FuenteDeEscurrimientos: boolean
      read flg_FuenteDeEscurrimientos write flg_FuenteDeEscurrimientos;
    property _flg_IgualPotenciaEnTodosLosPostes: boolean
      read flg_IgualPotenciaEnTodosLosPostes write flg_IgualPotenciaEnTodosLosPostes;
  {$ENDIF}

  end;

procedure THidroDePasada_cambioFichaPD(Actor: TCosa);

procedure AlInicio;
procedure AlFinal;

implementation


procedure THidroDePasada.dump_Variables(var f: TextFile; charIndentacion: char);
begin
  inherited dump_Variables(f, charIndentacion);
  writeln(f, charIndentacion, 'h_actual[m]= ', FloatToStrF(h_actual, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'salto_actual[m]= ', FloatToStrF(salto, ffFixed, 10, 3));

  writeln(f, charIndentacion, 'QAportePropio[m^3/s]= ',
    FloatToStrF(QAportePropio, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'VAportePropio[hm^3]= ',
    FloatToStrF(VAportePropio, ffFixed, 10, 3));

  writeln(f, charIndentacion, 'cv_USD_MWh[USD/MWh]= ',
    FloatToStrF(cv_USD_MWh, ffFixed, 10, 3));
  writeln(f);
end;

constructor THidroDePasada.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
  flg_CalcularGradienteDeInversion: boolean; fuenteDeAportes: TFuenteAleatoria;
  nombreBorne: string; flg_FuenteDeEscurrimientos, flg_IgualPotenciaEnTodosLosPostes:
  boolean; TonCO2xMWh: NReal; LowCostMustRun, CleanDevelopmentMechanism: boolean);

begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodo,
    TonCO2xMWh, LowCostMustRun, CleanDevelopmentMechanism,
    flg_CalcularGradienteDeInversion);
  Self.lpd := lpd;
  self.fuenteDeAportes := fuenteDeAportes;
  self.nombreBorne := nombreBorne;
  self.flg_FuenteDeEscurrimientos := flg_FuenteDeEscurrimientos;
  self.flg_IgualPotenciaEnTodosLosPostes := flg_IgualPotenciaEnTodosLosPostes;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor THidroDePasada.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);

  flg_FuenteDeEscurrimientos := False;
  flg_IgualPotenciaEnTodosLosPostes := False;

  f.IniciarLecturaRetrasada;
  f.rd('lpd', TCosa(lpd));
  f.rdReferencia('fuenteDeAportes', TCosa(fuenteDeAportes), Self);
  f.rd('nombreBorne', nombreBorne);
  if f.version > 104 then
    f.rd('flg_FuenteDeEscurrimientos', flg_FuenteDeEscurrimientos);
  if f.Version > 105 then
    f.rd('flg_IgualPotenciaEnTodosLosPostes', flg_IgualPotenciaEnTodosLosPostes);
  f.EjecutarLectura;
  pa := nil;
  nodo := nil;
  lpd.Propietario := self;

end;

{$ENDIF}

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure THidroDePasada.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('lpd', lpd);
  f.wrReferencia('fuenteDeAportes', fuenteDeAportes);
  f.wr('nombreBorne', nombreBorne);
  f.wr('flg_FuenteDeEscurrimientos', flg_FuenteDeEscurrimientos);
  f.wr('flg_IgualPotenciaEnTodosLosPostes', flg_IgualPotenciaEnTodosLosPostes);
end;

{$ENDIF}

procedure THidroDePasada.PrepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
  numeroBorne := fuenteDeAportes.idBorne(nombreBorne);
  cv_agua_USD_Hm3_Inc := 0;
  dh_RedQE := 0;
end;

procedure THidroDePasada.RegistrarParametrosDinamicos;
begin
  inherited registrarParametrosDinamicos;
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pA, nil,
    THidroDePasada_cambioFichaPD);
end;

function THidroDePasada.InfoAd_: string;
begin
  Result := '';
end;

function THidroDePasada.PotenciaFirme: NReal;
begin
  Result := (paUnidades.nUnidades_Operativas[0]) * pa.Pmax_Gen * pa.ren;
end;

class function THidroDePasada.DescClase: string;
begin
  Result := rsGHidraulicoPasada;
end;

procedure THidroDePasada.SorteosDelPaso(sortear: boolean);
begin
  if hayForzamientos or globs.ObligarDisponibilidad_1_ then
  begin
    NMaquinasDisponibles := paUnidades.nUnidades_Operativas[0];
    Pmax_Gen := pA.Pmax_gen;
    Qmax := pA.Qmax_Turb;
  end
  else if (sortear) then
  begin
    NMaquinasDisponibles := Sorteos_RepRotUnidades;
    Pmax_Gen := pA.Pmax_gen;
    Qmax := pA.Qmax_Turb;
  end
  else
  begin
    NMaquinasDisponibles := paUnidades.nUnidades_Operativas[0];
    Pmax_Gen := pA.Pmax_gen * pA.fdispo;
    Qmax := pA.Qmax_Turb * pA.fdispo;
  end;
  PDisponible_Central := NMaquinasDisponibles * Pmax_Gen;
end;

function THidroDePasada.ReduccionDeCotaPorCaudalErogado(QErogado: NReal): NReal;
begin
  //Si no tengo centrales aguas arriba se que lo que voy a erogar son todos mis
  //aportes
  if self.pa.centralesAguasArriba.Count = 0 then
    dh_RedQE := (pa.cbQE * QAportePropio + pa.caQE) * QAportePropio
  else
    dh_RedQE := (pa.cbQE * QErogado + pa.caQE) * QErogado;

  Result := dh_RedQE;
end;

function THidroDePasada.centralLagoDescarga: TGeneradorHidraulico;
begin
  Result := pa.central_lagoDescarga;
end;

function THidroDePasada.vertimientoPorPoste: boolean;
begin
  Result := False;
end;

function THidroDePasada.NVariablesPotenica: integer;
begin
  if flg_IgualPotenciaEnTodosLosPostes then
    Result := 1
  else
    Result := globs.NPostes;
end;


{$IFDEF MIGRACION_PERSISTENCIA}
class function THidroDePasada.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: integer): TDataColumnListOfCosa;
begin
  Result := inherited CreateDataColumnList(xClaseDeCosa, xVersion);
  Result.AddCosaColumn('fuenteDeAportes');
  Result.AddStringColumn('nombreBorne', '');
  Result.AddBooleanColumn('flg_FuenteDeEscurrimientos', False, 105);
  Result.AddBooleanColumn('flg_IgualPotenciaEnTodosLosPostes', False, 106);
end;

{$ENDIF}

procedure THidroDePasada.AfterInstantiation;
begin
  inherited AfterInstantiation;
  pa := nil;
end;



procedure THidroDePasada.ResolverEncadenamientos(s: TSimplex);
var
  iCentrales: integer;
  gh: TFichaCentralAguasArriba;
begin
  // Aportes de las centrales aguas arriba
  // sumamos los aportes de las centrales aguas arriba
  // sobre la restricción de vertimiento
  for iCentrales := 0 to pa.centralesAguasArriba.Count - 1 do
  begin
    gh := TFichaCentralAguasArriba(pa.centralesAguasArriba[iCentrales]);
    gh.Central.CargarAportesAFilaCentralAguasAbajo(
      ires, // fila en la que cargar los aportes
      gh.coef, // coeficiente de llegada por el que multiplicar los aportes
      pa.cv_agua_USD_Hm3,   // valor del agua a considerar en la central aguas abajo.
      s);
  end;
end;


function THidroDePasada.CotaAguasArriba: NReal;
begin
  Result := pa.hToma;
end;

function THidroDePasada.CotaDescarga: NReal;
begin
  if pa.central_lagoDescarga = nil then
    Result := pa.hDescarga
  else
    Result := pa.central_LagoDescarga.cotaAguasArriba;
end;

procedure THidroDePasada.PrepararPaso_ps;
begin
  h_actual := pa.hToma;
  QAportePropio := Self.fuenteDeAportes.Bornera[Self.numeroBorne];

  if flg_FuenteDeEscurrimientos then
    QAportePropio := QAportePropio * fEscurrimientoToCaudal;


(*
    Assert(QAportePropio >= 0, 'El generador hidráulico de pasada ' + nombre +
    ' tiene caudal de aportes propios negativo. Fuente: ' + fuenteDeAportes.nombre
    + ', Borne: ' + nombreBorne);
*)
  if QAportePropio < 0 then
    QAportePropio := 0;



  VAportePropio := QAportePropio * globs.SegundosDelPaso / 1E6;
  Salto := max(0.1, h_actual - CotaDescarga - ReduccionDeCotaPorCaudalErogado(
    QErogado_));
  ce := Salto * ro_g_ren;

  cv_USD_MWh := pa.cv_agua_USD_Hm3 / (1.0E6 * ce) * globs.fActPaso;
  PmaxParaQmax := Qmax * ce * 3600;
  Pmax_Central := NMaquinasDisponibles * min(Pmax_Gen, PmaxParaQmax);
  PuedoGenerar := (NMaquinasDisponibles > 0) and (Salto >= pa.SaltoMinimoOperativo);
end;

procedure THidroDePasada.opt_nvers(var ivar, ivae, ires: integer);
begin
  self.ivar := ivar;
  self.ires := ires;

  if not flg_IgualPotenciaEnTodosLosPostes then
    ivar := ivar + globs.NPostes + 1//Potencias por poste y Vertimiento
  else
    ivar := ivar + 2; // Una Potenica y el Vertimiento.

  Inc(ires);// -Turbinado -Vertimiento + Aportes = 0
  // ponemos una restricción que equivale a decir
  // que todos los Aportes son Erogados, ya sea como
  // Turbinado o Vertimiento

  if pA.HayRestriccionEmaxPasoDeTiempo then
    Inc(ires);
end;

procedure THidroDePasada.opt_cargue(s: TSimplex);
var
  ibaseres, ires_: integer;
  iposte: integer;
begin
{$IFDEF SPXCONLOG}
  spx_NombrarVariables(s);
{$ENDIF}
  ibaseres := nodo.ires;
  // Tenemos que aportar a las restricciones de demanda del nodo al que está
  // conectado el generador

  if not flg_IgualPotenciaEnTodosLosPostes then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(ibaseres + iposte, ivar + iposte, 1);
    ivert := ivar + globs.NPostes;
    //Restricción -Turbinado -Vertimiento + Aportes = 0
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(ires, ivar + iposte, -globs.durpos[iposte] / (ce * 1.0E6));

    // Ahora agregamos a la función de UTILIDAD (-costo)
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(s.nf, ivar + iposte, -cv_USD_MWh * globs.durpos[iposte]);

  end
  else
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(ibaseres + iposte, ivar, 1);
    ivert := ivar + 1;
    //Restricción -Turbinado -Vertimiento + Aportes = 0
    s.pon_e(ires, ivar, -globs.HorasDelPaso / (ce * 1.0E6));

    // Ahora agregamos a la función de UTILIDAD (-costo)
    s.pon_e(s.nf, ivar, -cv_USD_MWh * globs.HorasDelPaso);

  end;

  s.pon_e(ires, ivert, -1); // -Vertimiento
  s.pon_e(ires, s.nc, VAportePropio);
  s.FijarRestriccionIgualdad(ires);

  s.pon_e(s.nf, ivert, -pa.cv_agua_USD_Hm3);


  ires_ := ires;
  //Restriccion a la energía máxima generable
  if pA.HayRestriccionEmaxPasoDeTiempo then
  begin
    Inc(ires_);
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(ires_, ivar + iposte, -globs.durpos[iposte]);
    s.pon_e(ires_, s.nc, pA.EmaxPasoDeTiempo);
  end;

end;

procedure THidroDePasada.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  iposte: integer;
  NPs: integer;
begin

  if not flg_IgualPotenciaEnTodosLosPostes then
    NPs := globs.NPostes
  else
    NPs := 1;

  // Le fijamos como cota máxima  de la potencia en cada poste al mínimo
  // entre la PMax del generador y lo que puede generar con el caudal máximo
  // para el ce
  if PuedoGenerar then
  begin
    if (not hayForzamientos) then
    begin
      for iposte := 0 to NPs - 1 do
        s.cota_sup_set(ivar + iposte, PMax_Central);
    end
    else
    begin
      for iposte := 0 to NPs - 1 do
        s.FijarVariable(ivar + iposte, paForzamiento.P[0]);
    end;
  end
  else
  begin
    for iposte := 0 to NPs - 1 do
      s.FijarVariable(ivar + iposte, 0);
  end;
  s.cota_sup_set(ivert, 9000000); //Qmax * globs.SegundosDelPaso / 1E2);
end;

procedure THidroDePasada.Sim_Cronica_Inicio;
begin
  inherited Sim_Cronica_Inicio;
  QErogado_ := 0;

  QErogado_IterAnterior := 0; // para que de igual al correrlo dos veces seguidas

  VVertido := 0;
  VTurbinado := 0;
  VErogado := 0;
  EnergiaGenerada := 0;
end;

procedure THidroDePasada.opt_leerSolucion(s: TSimplex);
var
  iposte: integer;
begin

  if PuedoGenerar then
  begin
    // recuperamos los valores de Potencia despachada
    EnergiaGenerada := 0;

    if not flg_IgualPotenciaEnTodosLosPostes then
    begin
      for iposte := 0 to globs.NPostes - 1 do
      begin
        P[iposte] := s.xval(ivar + iposte);
        EnergiaGenerada := EnergiaGenerada + P[iposte] * globs.durpos[iposte];
      end;
    end
    else
    begin
      P[0] := s.xval(ivar);
      EnergiaGenerada := P[0] * globs.durpos[iposte];
      for iposte := 1 to globs.NPostes - 1 do
      begin
        P[iposte] := P[0];
        EnergiaGenerada := EnergiaGenerada + P[iposte] * globs.durpos[iposte];
      end;
    end;
  end
  else
  begin
    for iposte := 0 to globs.NPostes - 1 do
      P[iposte] := 0;
    EnergiaGenerada := 0;
  end;



  Ingreso_PorDisponibilidad :=
    PDisponible_Central * pa.PagoPorDisponibilidad_USD_MWh * globs.HorasDelPaso;
  Ingreso_PorEnergia := EnergiaGenerada * pa.PagoPorEnergia_USD_MWh;
  costoDirectoDelPaso := Ingreso_PorEnergia + Ingreso_PorDisponibilidad;

  cv_agua_USD_Hm3_MLRB := -s.ymult(ires);

  VVertido := s.xval(ivert);
  Dual_Vertimiento := -s.xmult(ivert);

  VTurbinado := EnergiaGenerada / (ce * 1.0E6);
  VErogado := VTurbinado + VVertido;

  QVertido := VVertido * (1.0E6 * globs.invSegundosDelPaso);
  QTurbinado := VTurbinado * (1.0E6 * globs.invSegundosDelPaso);
  QErogado_ := VErogado * (1.0E6 * globs.invSegundosDelPaso);
end;

{$IFDEF SPXCONLOG}
procedure THidroDePasada.spx_NombrarVariables(s: TSimplex);
var
  iposte: integer;
begin
  for iposte := 0 to globs.NPostes - 1 do
    s.set_NombreVar(ivar + iposte, Nombre + '_P[MW]' + IntToStr(iposte + 1));
  s.set_NombreVar(ivar + globs.NPostes, Nombre + 'Vert[Hm3]');

  s.set_NombreRest(ires, Nombre + '_-VErogado+VAportes=0');

  if pA.HayRestriccionEmaxPasoDeTiempo then
    s.set_NombreRest(ires + 1, Nombre + '_Emax');
end;

{$ENDIF}



function THidroDePasada.opt_necesitoIterar(kIteracion: integer;
  var errRelativo: NReal): boolean;
var
  qm: NReal;
  qz: NReal;
  alfa: NReal;
begin
  qm := max(QErogado_, QErogado_IterAnterior);
  if qm > 1 then
    errRelativo := abs(QErogado_ - QErogado_IterAnterior) / qm
  else
    errRelativo := 0;


  alfa := 0.7;
  QErogado_ := alfa * QErogado_IterAnterior + (1 - alfa) * QErogado_;

  qz := QErogado_;
  QErogado_IterAnterior := qz;

  Result := errRelativo > 0.1;
end;



function THidroDePasada.getNombreVar(ivar: integer; var nombre: string): boolean;
begin
  if (ivar >= self.ivar) and (ivar < self.ivar + globs.NPostes) then
  begin
    nombre := self.Nombre + '_P[MW]' + IntToStr(ivar - self.ivar + 1);
    Result := True;
  end
  else if (ivar = ivert) then
  begin
    nombre := self.Nombre + '_Vertimiento[Hm3]';
    Result := True;
  end
  else
    Result := False;
end;

function THidroDePasada.getNombreRes(ires: integer; var nombre: string): boolean;
begin
  if (ires = self.ires) then
  begin
    nombre := self.nombre + '_-VErogado+VAportes=0';
    Result := True;
  end
  else if pa.HayRestriccionEmaxPasoDeTiempo and (ires = self.ires + 1) then
  begin
    nombre := self.nombre + '_res-EMax';
    Result := True;
  end
  else
    Result := False;
end;

procedure THidroDePasada.PubliVars;
begin
  inherited PubliVars;
  PublicarVariableNR('QAportesP', '[m3/s]', 8, 1, QAportePropio, True);
  PublicarVariableNR('QTurbinado', '[m3/s]', 8, 1, QTurbinado, True);
  PublicarVariableNR('QVertido', '[m3/s]', 8, 1, QVertido, True);
  PublicarVariableNR('Salto', '[m]', 6, 1, Salto, True);
  PublicarVariableNR('ce', '[MWh/m3]', 12, 6, ce, True);
  PublicarVariableNR('dh_RedQE', '[m]', 6, 2, dh_RedQE, True);
  PublicarVariableNR('cv_USD_MWh', '[USD/MWh]', 6, 1, cv_USD_MWh, True);
  PublicarVariableNR('cv_agua_MLRB', '[USD/Hm3]', 6, 1, cv_agua_USD_Hm3_MLRB, True);
  PublicarVariableNR('Cota', '[m]', 6, 1, h_actual, False);
  PublicarVariableNR('VVertido', '[Hm3]', 8, 1, VVertido, False);
  PublicarVariableNR('Dual_Vertimiento', '[USD/Hm3]', 8, 1, Dual_Vertimiento, True);
  PublicarVariableNR('Pmax_Central', '[MW]', 8, 1, Pmax_Central, True);

  PublicarVariableNR('VTurbinado', '[Hm3]', 8, 1, VTurbinado, False);
  PublicarVariableNI('NMaqsDisponibles', '-', NMaquinasDisponibles, False);

end;

procedure THidroDePasada.optx_nvxs(var ixr, ixd, iauxr, iauxd: integer);
begin
  Self.iauxr := iauxr;
  iauxr := iauxr + 1;
end;

procedure THidroDePasada.PosicionarseEnEstrellita;
begin
  Self.QErogado_ := globs.Auxs_r0[iauxr][globs.CF.ordinalEstrellaActual];

  //Para que tenga el mismo estado de iteraciones al posicionarse en la estrella y
  //de igual en la optimización monohilo y multihilo
  Self.QErogado_IterAnterior := QErogado_;

end;

procedure THidroDePasada.AcumAux1(peso: NReal);
begin
  globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] :=
    globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] + Self.QErogado_ * peso;
end;

procedure THidroDePasada.SetAux1;
begin
  globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] := Self.QErogado_;
end;

procedure THidroDePasada.Free;
begin
  if lpd <> nil then
    lpd.Free;
  inherited Free;
end;


function THidroDePasada.get_pa_FD(kTipoUnidad: integer): NReal;
begin
  Result := pa.fDispo;
end;

function THidroDePasada.get_pa_TMR(kTipoUnidad: integer): NReal;
begin
  Result := pa.tRepHoras;
end;


procedure THidroDePasada.CambioFichaPD;
begin
  ActualizarProbabilidadesReparacionYRotura_(pa.fDispo, pa.tRepHoras);
  ro_g_ren := pA.ren * ro_Agua * g_gravitacion / (1.0E6 * 3600.0);
  if flg_FuenteDeEscurrimientos then
    fEscurrimientoToCaudal := 1.0 / 1000 * pa.AreaCuenca_ha * 100 * 100 / (730 * 3600);

end;

{*********************************
*Métodos de TFichaHidroDePasada*
*********************************}

constructor TFichaHidroDePasada.Create(capa: integer; fecha: TFecha;
  periodicidad: TPeriodicidad; AreaCuenca_ha, hDescarga, hToma: NReal;
  cv_agua_USD_Hm3: NReal; central_lagoDescarga: TGeneradorHidraulico;
  centralesAguasArriba: TListaCentralesAguasArriba;
  caQE, cbQE, ren_, Pmax_Gen_, Qmax_Turb_: NReal; SaltoMinimoOperativo: NReal;
  fDispo_: NReal; tRepHoras: NReal; HayRestriccionEmaxPasoDeTiempo: boolean;
  EmaxPasoDeTiempo: NReal; PagoPorDisponibilidad_USD_MWh: NReal;
  PagoPorEnergia_USD_MWh: NReal);
begin
  inherited Create(capa, fecha, periodicidad);
  self.AreaCuenca_ha := AreaCuenca_ha;
  self.hDescarga := hDescarga;
  self.hToma := hToma;
  self.cv_agua_USD_Hm3 := cv_agua_USD_Hm3;
  self.central_lagoDescarga := central_lagoDescarga;
  self.caQE := caQE;
  self.cbQE := cbQE;
  self.ren := ren_;
  self.Pmax_Gen := Pmax_Gen_;
  self.Qmax_Turb := Qmax_Turb_;
  self.SaltoMinimoOperativo := SaltoMinimoOperativo;
  self.fDispo := fDispo_;
  self.tRepHoras := tRepHoras;
  self.centralesAguasArriba := centralesAguasArriba;
  self.HayRestriccionEmaxPasoDeTiempo := HayRestriccionEmaxPasoDeTiempo;
  self.SaltoMinimoOperativo := saltoMinimoOperativo;
  self.EmaxPasoDeTiempo := EmaxPasoDeTiempo;
  self.PagoPorDisponibilidad_USD_MWh := PagoPorDisponibilidad_USD_MWh;
  self.PagoPorEnergia_USD_MWh := PagoPorEnergia_USD_MWh;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure TFichaHidroDePasada.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('AreaCuenca_ha', AreaCuenca_ha);
  f.wr('hDescarga', hDescarga, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('cv_agua_USD_Hm3', cv_agua_USD_Hm3, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('hToma', hToma, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('caQE', caQE, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('cbQE', cbQE, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('ren', ren, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('Pmax_Gen', Pmax_Gen, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('Qmax_Turb', Qmax_Turb, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('SaltoMinimoOperativo', saltoMinimoOperativo);
  f.wr('fDispo', fDispo, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('tRepHoras', tRepHoras, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALESPU);
  f.wrReferencia('central_lagoDescarga', central_lagoDescarga);
  f.wr('centralesAguasArriba', centralesAguasArriba);
  f.wr('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
  f.wr('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
  f.wr('PagoPorDisponibilidad_USD_MWh', PagoPorDisponibilidad_USD_MWh);
  f.wr('PagoPorEnergia_USD_MWh', PagoPorEnergia_USD_MWh);
end;

{$ENDIF}

function TFichaHidroDePasada.infoAd_: string;
begin
  Result := 'PMáxGen= ' + FloatToStrF(Pmax_Gen, ffGeneral, 10, 1) +
    ' MW, ' + 'QMáxTurb= ' + FloatToStrF(Qmax_Turb, ffGeneral, 10, 1) +
    ' m^3/s, ' + 'ren= ' + FloatToStrF(ren, ffGeneral, 10, 2) +
    ' p.u., ' + 'fDisp= ' + FloatToStrF(fDispo, ffGeneral, 10, 2) +
    ' p.u., ' + 'tRep= ' + FloatToStrF(tRepHoras, ffGeneral, 10, 1) +
    'h, ' + 'hToma= ' + FloatToStrF(hToma, ffGeneral, 10, 2) +
    ' m, ' + 'hDescarga= ' + FloatToStrF(hDescarga, ffGeneral, 10, 2) + 'm';
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TFichaHidroDePasada.Create_ReadFromText(f: TArchiTexto);
var
  cv_agua_USD_MWh: NReal;
begin

  PagoPorDisponibilidad_USD_MWh := 0.0;
  PagoPorEnergia_USD_MWh := 0.0;
  AreaCuenca_ha := 0;

  if f.Version < 35 then
  begin
    if f.Version < 14 then
    begin
      inherited Create_ReadFromText(f);
      f.IniciarLecturaRetrasada;
      f.rd('hDescarga', hDescarga);
      f.rd('cv_agua_USD_MWh', cv_agua_USD_MWh);
      f.rd('hToma', hToma);
      f.rd('caQE', caQE);
      f.rd('cbQE', cbQE);
      f.rd('ren', ren);
      f.rd('Pmax_Gen', Pmax_Gen);
      f.rd('Qmax_Turb', Qmax_Turb);
      f.rd('fDispo', fDispo);
      f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
      f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
      f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
      f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
      f.EjecutarLectura;
      self.tRepHoras := 48;
      self.SaltoMinimoOperativo := 0.1;
    end
    else
    begin
      inherited Create_ReadFromText(f);
      f.IniciarLecturaRetrasada;
      f.rd('hDescarga', hDescarga);
      f.rd('cv_agua_USD_MWh', cv_agua_USD_MWh);
      f.rd('hToma', hToma);
      f.rd('caQE', caQE);
      f.rd('cbQE', cbQE);
      f.rd('ren', ren);
      f.rd('Pmax_Gen', Pmax_Gen);
      f.rd('Qmax_Turb', Qmax_Turb);
      if f.Version >= 30 then
        f.rd('SaltoMinimoOperativo', saltoMinimoOperativo)
      else
        self.SaltoMinimoOperativo := 0.1;
      f.rd('fDispo', fDispo);
      f.rd('tRepHoras', tRepHoras);
      f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
      f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
      f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
      f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
      f.EjecutarLectura;
    end;
    cv_agua_USD_Hm3 := 0;
  end
  else
  begin
    inherited Create_ReadFromText(f);
    f.IniciarLecturaRetrasada;
    if f.Version > 104 then
      f.rd('AreaCuenca_ha', AreaCuenca_ha);
    f.rd('hDescarga', hDescarga);
    f.rd('cv_agua_USD_Hm3', cv_agua_USD_Hm3);
    f.rd('hToma', hToma);
    f.rd('caQE', caQE);
    f.rd('cbQE', cbQE);
    f.rd('ren', ren);
    f.rd('Pmax_Gen', Pmax_Gen);
    f.rd('Qmax_Turb', Qmax_Turb);
    f.rd('SaltoMinimoOperativo', saltoMinimoOperativo);
    f.rd('fDispo', fDispo);
    f.rd('tRepHoras', tRepHoras);
    f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
    f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
    f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
    f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
  end;
  if f.Version >= 98 then
  begin
    f.rd('PagoPorDisponibilidad_USD_MWh', PagoPorDisponibilidad_USD_MWh);
    f.rd('PagoPorEnergia_USD_MWh', PagoPorEnergia_USD_MWh);
  end;
  f.EjecutarLectura;
end;

{$ENDIF}

procedure TFichaHidroDePasada.Free;
begin
  if centralesAguasArriba <> nil then
    centralesAguasArriba.Free;
  inherited Free;
end;

{$IFDEF MIGRACION_PERSISTENCIA}
class function TFichaHidroDePasada.CreateDataColumnList(xClaseDeCosa: TClaseDeCosa;
  xVersion: integer): TDataColumnListOfCosa;
begin
  Result := inherited CreateDataColumnList(xClaseDeCosa, xVersion);
  Result.AddDoubleColumn('AreaCuenca_ha', 0.0, 105);
  Result.AddDoubleColumn('hDescarga', 0.0);
  Result.AddDoubleColumn('cv_agua_USD_MWh', 0.0, 0, 35);
  Result.AddDoubleColumn('cv_agua_USD_Hm3', 0.0, 35);
  Result.AddDoubleColumn('hToma', 0.0);
  Result.AddDoubleColumn('caQE', 0.0);
  Result.AddDoubleColumn('cbQE', 0.0);
  Result.AddDoubleColumn('ren', 0.0);
  Result.AddDoubleColumn('Pmax_Gen', 0.0);
  Result.AddDoubleColumn('Qmax_Turb', 0.0);
  Result.AddDoubleColumn('SaltoMinimoOperativo', 0.1, 30);
  Result.AddDoubleColumn('fDispo', 0.0);
  Result.AddDoubleColumn('tRepHoras', 48, 14);
  Result.AddCosaColumn('central_lagoDescarga');
  Result.AddCosaColumn('centralesAguasArriba');
  Result.AddBooleanColumn('HayRestriccionEmaxPasoDeTiempo', False);
  Result.AddDoubleColumn('EmaxPasoDeTiempo', 0.0);
  Result.AddDoubleColumn('PagoPorDisponibilidad_USD_MWh', 0.0, 98);
  Result.AddDoubleColumn('PagoPorEnergia_USD_MWh', 0.0, 98);
end;

{$ENDIF}



procedure THidroDePasada_cambioFichaPD(Actor: TCosa);
begin
  (Actor as THidroDePasada).cambioFichaPD;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(THidroDePasada.ClassName, THidroDePasada);
  registrarClaseDeCosa(TFichaHidroDePasada.ClassName, TFichaHidroDePasada);
end;

procedure AlFinal;
begin
end;

end.
